obj22 <- function(A, S, B, response, gamma){  # objective function with only S(graphical part), without L(IRT part)
  S_temp <- S;
  diag(S_temp) <- 0;
  neg_loglik_cpp22(A, S, B, response) + gamma * sum(abs(S_temp));
}

grad_A22 <- function(A0, S0, B0, Q, response){#Calculate the gradient w.r.t. A.
  L0 <- A0 %*% B0 %*% t(B0) %*% t(A0)
  temp <- neg_loglik_dev(L0, S0, response)
  temp_L <- 2 * temp %*% A0 %*% B0 %*% t(B0)
  temp_L*Q;
}

Update_A22 <- function(A0, S0, B0, Q, response, step = 1, disc = 0.5){
  grad = grad_A_cpp22(A0, S0, B0, Q, response)
  neg_loglik0 <- neg_loglik_cpp22(A0, S0, B0, response)

  A <- A0 - step * grad;
  neg_loglik1 <- neg_loglik_cpp22(A, S0, B0, response)

  while(neg_loglik1 > neg_loglik0){ #Search for a step size such that the objective function increases.
    step <- step * disc;
    A <- A0 - step * grad;
    neg_loglik1 <- neg_loglik_cpp22(A, S0, B0, response)
  }
  list(step = step, A = A);
}

Update_B22 <- function(A0, S0, B0, response, step = 1, disc = 0.5){
  grad <- grad_B_cpp22(A0, S0, B0, response)
  grad[upper.tri(grad,diag = T)] <- 0
  neg_loglik0 <- neg_loglik_cpp22(A0, S0, B0, response)

  B <- B0 - step * grad
  neg_loglik1 <- neg_loglik_cpp22(A0, S0, B, response)
  while(neg_loglik1 > neg_loglik0){
    step <- step * disc
    B <- B0 - step * grad
    neg_loglik1 <- neg_loglik_cpp22(A0, S0, B, response)
  }
  list(step = step, B = B)
}

fhat22 <- function(S, S0, A0, B0, response, step){ #Use to decide when to stop line searching.
  grad <- grad_S_cpp22(A0, S0, B0, response);
  neg_loglik_cpp22(A0, S0, B0, response) + sum(grad * (S - S0)) + sum((S - S0)^2)/(2*step);
}

Update_S22 <- function(A0, S0, B0, gamma, response, step = 1, disc = 0.5){
  grad <- grad_S_cpp22(A0, S0, B0, response)
  y <- S0 - step * grad;
  temp <- soft_thre_cpp(y, gamma*step);
  diag(temp) <- diag(y);
  S <- temp;

  neg_loglik1 <- neg_loglik_cpp22(A0, S, B0, response);
  fhat1 <- fhat22(S, S0, A0, B0, response, step)

  while(neg_loglik1 > fhat1){
    step = step * disc;

    y <- S0 - step * grad;
    temp <- soft_thre_cpp(y, gamma*step);
    diag(temp) <- diag(y);
    S <- temp;

    neg_loglik1 <- neg_loglik_cpp22(A0, S,B0, response);
    fhat1 <- fhat22(S, S0, A0, B0, response, step)
  }
  list(step = step, S = S)
}


#' @export confFLaG
confFLaG <- function(A0, S0, Sigma0, Q, gamma, response, stepA = 5, stepS = 5, stepB = 5, discA = 0.5,
                     discS = 0.5, discB = 0.5, tol = 1e-5, print_proc = T){
  if(is.null(dim(A0))||is.data.frame(A0))
    A0 <- as.matrix(A0)
  if(is.null(dim(S0))||is.data.frame(S0))
    S0 <- as.matrix(S0)
  if(is.null(dim(Sigma0))||is.data.frame(Sigma0))
    Sigma0 <- as.matrix(Sigma0)
  if(is.null(dim(Q))||is.data.frame(Q))
    Q <- as.matrix(Q)
  if(is.null(dim(response))||is.data.frame(response))
    response <- as.matrix(response)
  B0 <- t(chol(Sigma0))
  a <- diag(B0)
  B0 <- diag(1/a) %*% B0
  A0 <- A0 %*% diag(a)

  obj0 <- obj22(A0, S0, B0, response, gamma)

  resultA <- Update_A22(A0, S0, B0, Q, response, stepA, discA)
  stepA <- resultA$step
  A <- resultA$A;
  resultS <- Update_S22(A, S0, B0, gamma, response, stepS, discS);
  stepS <- resultS$step
  S <- resultS$S
  resultB <- Update_B22(A, S, B0, response, stepB, discB)
  stepB <- resultB$step
  B <- resultB$B

  obj1 <- obj22(A, S, B, response, gamma)
  k <- 1
  thre_old <- 1e3
  while(obj0 - obj1 > tol){
    if(print_proc == T){
      thre <- min(obj0-obj1,thre_old)
      thre_old <- thre
      nn <- max(round(30*(log(thre)/log(tol))),0)
      cat(paste0("\r|",strrep("=",nn)),strrep(" ",30-nn),"\b\b|",round(100*(log(thre)/log(tol))),"\b%", " Iteration: ", k, " thre = ",thre)
    }
    A0 <- A
    S0 <- S
    B0 <- B
    obj0 <- obj1

    resultA <- Update_A22(A0, S0, B0, Q, response, stepA, discA)
    stepA <- resultA$step
    A <- resultA$A
    resultS <- Update_S22(A, S0, B0, gamma, response, stepS, discS)
    stepS <- resultS$step
    S <- resultS$S
    resultB <- Update_B22(A, S, B0, response, stepB)
    stepB <- resultB$step
    B <- resultB$B
    obj1 <- obj22(A, S, B, response, gamma)
    k <- k + 1
    #    print(obj1);
  }
  Sigma <- B %*% t(B)
  a <- sqrt(diag(Sigma))
  A_a <- A
  A <- A %*% diag(a)
  Sigma <- diag(1/a) %*% Sigma %*% diag(1/a)
  list(A = A, S = S, Sigma = Sigma)
}

#' @export confFLaG.supp
confFLaG.supp <- function(A0, S0, Sigma0, Q, supp, response, stepA = 1, stepS = 1, stepB = 1,
                          discA = 0.5, discS = 0.5, discB = 0.5, tol = 1e-5, print_proc = T){
  if(is.null(dim(A0))||is.data.frame(A0))
    A0 <- as.matrix(A0)
  if(is.null(dim(S0))||is.data.frame(S0))
    S0 <- as.matrix(S0)
  if(is.null(dim(Sigma0))||is.data.frame(Sigma0))
    Sigma0 <- as.matrix(Sigma0)
  if(is.null(dim(Q))||is.data.frame(Q))
    Q <- as.matrix(Q)
  if(is.null(dim(response))||is.data.frame(response))
    response <- as.matrix(response)
  B0 <- t(chol(Sigma0))
  a <- diag(B0)
  B0 <- diag(1/a) %*% B0
  A0 <- A0 %*% diag(a)

  obj0 <- neg_loglik_cpp22(A0, S0, B0, response)

  resultA <- Update_A22(A0, S0, B0, Q, response, stepA, discA);
  stepA <- resultA$step;
  A <- resultA$A;
  resultS <- Update_S_supp_cpp22(A, S0, B0, supp, response, stepS, discS);
  stepS <- resultS$step;
  S <- resultS$S;
  resultB <- Update_B22(A, S, B0, response, stepB, discB)
  B <- resultB$B
  stepB <- resultB$step
  obj1 <- neg_loglik_cpp22(A, S, B, response);
  k <- 1
  thre_old <- 1e3
  while(obj0 - obj1 > tol){
    if(print_proc == T){
      thre <- min(obj0-obj1,thre_old)
      thre_old <- thre
      nn <- max(round(30*(log(thre)/log(tol))),0)
      cat(paste0("\r|",strrep("=",nn)),strrep(" ",30-nn),"\b\b|",round(100*(log(thre)/log(tol))),"\b%", " Iteration: ", k, " thre = ",thre)
    }

    A0 <- A
    S0 <- S
    B0 <- B
    obj0 <- obj1

    resultA <- Update_A22(A0, S0, B0, Q, response, stepA, discA)
    stepA <- resultA$step;
    A <- resultA$A
    resultS <- Update_S_supp_cpp22(A, S0, B0, supp, response, stepS, discS)
    stepS <- resultS$step;
    S <- resultS$S
    resultB <- Update_B22(A, S, B0, response, stepB, discB)
    B <- resultB$B
    stepB <- resultB$step
    obj1 <- neg_loglik_cpp22(A, S, B, response)
    k <- k + 1
    #print(obj1);
  }
  Sigma <- B %*% t(B)
  a <- sqrt(diag(Sigma))
  A_a <- A
  A <- A %*% diag(a)
  Sigma <- diag(1/a) %*% Sigma %*% diag(1/a)
  list(A = A, S = S*supp, Sigma = Sigma,
       pseudoLoglik = -nrow(response)*neg_loglik_cpp22(A_a, S, B, response),
       scoreEst = response %*%  A %*%  Sigma)
}
#' @export FLaG.sample
FLaG.sample <- function(A, S, Sigma, N, burn = 200){
  return(FLaG_sample(A=A,S=S,Sigma = Sigma,N=N, burn = burn))
}
