#include "depend_functions.h"
#include "lbfgs.h"
#include "progress.hpp"
#include <stdio.h>
#include <stdlib.h>

static lbfgsfloatval_t evaluate22(
    void *instance,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step
)
{
  int i,j,k=0;

  lbfgsfloatval_t fx = 0.0;
  lbfgsfloatval_t *params = (lbfgsfloatval_t *)instance;
  const int N = params[k++];
  const int J = params[k++];
  const int K = params[k++];

  arma::mat A = arma::zeros(J,K);
  arma::vec b(J);
  arma::mat response(N,J);
  arma::mat Q = arma::zeros(J, K);
  arma::mat B = arma::eye(K,K);
  for(i=0;i<N;++i){
    for(j=0;j<J;++j){
      response(i,j) = params[k++];
    }
  }
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      Q(i,j) = params[k++];
    }
  }

  k=0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
        A(i,j) = x[k++];
    }
  }
  for(i=0;i<J;++i){
    b(i) = x[k++];
  }
  for(i=1;i<K;++i){
    for(j=0;j<i;++j){
      B(i,j) = x[k++];
    }
  }
  arma::mat S = arma::diagmat(b);
  fx = neg_loglik_cpp22(A, S, B, response) * N;

  arma::mat g_A = grad_A_cpp22(A, S, B, Q, response);
  arma::mat g_S = grad_S_cpp22(A, S, B,  response);
  arma::mat g_B = grad_B_cpp22(A, S, B, response);

  k=0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      g[k++] = g_A(i,j);
    }
  }
  for(i=0;i<J;++i){
    g[k++] = g_S(i,i);
  }
  for(i=1;i<K;++i){
    for(j=0;j<i;++j){
      g[k++] = g_B(i,j);
    }
  }
  return fx;
}

static int progress22(
    void *instance,
    const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx,
    const double rate,
    const lbfgsfloatval_t delta,
    const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step,
    int n,
    int k,
    int ls
)
{
  Progress p(0, false);
  if(Progress::check_abort())
    return 2;

  int tt = k;
  if((tt%4)==0){
    printf("\r- Iteration %d:", k);
    printf(" obj = %.1f, obj_change = %.1f, ", fx, rate);
  }
  else if((tt%4)==1){
    printf("\r\\ Iteration %d:", k);
    printf(" obj = %.1f, obj_change = %.1f, ", fx, rate);
  }
  else if((tt%4)==2){
    printf("\r| Iteration %d:", k);
    printf(" obj = %.1f, obj_change = %.1f, ", fx, rate);
  }
  else{
    printf("\r/ Iteration %d:", k);
    printf(" obj = %.1f, obj_change = %.1f, ", fx, rate);
  }

  return 0;
}

// [[Rcpp::export]]
Rcpp::List AM_IRT_cpp22(arma::mat A0, arma::vec b0, arma::mat B0, arma::mat response,
                       arma::mat Q, int print_proc,
                       double accuracy, int min_step, int max_step){
  int ret = 0;
  const int N = response.n_rows;
  const int J = response.n_cols;
  const int K = A0.n_cols;
  int i,j=0;

  lbfgsfloatval_t fx=0;
  lbfgsfloatval_t *x = lbfgs_malloc(J*(K+1)+K*(K-1)/2);
  lbfgs_parameter_t param;
  lbfgs_parameter_init(&param);
  param.delta = accuracy;
  param.past = min_step;
  param.max_iterations = max_step;
  //  printf("epsilon=%f\n", param.epsilon);
  lbfgsfloatval_t *params = lbfgs_malloc(N*J+J*K+20);
  /* Initialize the constant params. */
  int k=0;
  params[k++] = N;
  params[k++] = J;
  params[k++] = K;

  for(i=0;i<N;++i){
    for(j=0;j<J;++j){
      params[k++] = response(i,j);
    }
  }
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      params[k++] = Q(i,j);
    }
  }

  /* Initialize the variables. */
  k=0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      x[k++] = A0(i,j);
    }
  }
  for(i=0;i<J;++i){
    x[k++] = b0(i);
  }
  for(i=1;i<K;++i){
    for(j=0;j<i;++j){
      x[k++] = B0(i,j);
    }
  }
  printf("Press Esc to interrupt\n\n");
  printf("Accuracy = %f\n", accuracy);
  if(print_proc)
  {
    ret = lbfgs(J*(K+1)+K*(K-1)/2, x, &fx, evaluate22, progress22, params, &param);
  }
  else
    ret = lbfgs(J*(K+1)+K*(K-1)/2, x, &fx, evaluate22, NULL, params, &param);

  k=0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      A0(i,j) = x[k++];
    }
  }
  for(i=0;i<J;++i){
    b0(i) = x[k++];
  }
  for(i=1;i<K;++i){
    for(j=0;j<i;++j){
      B0(i,j) = x[k++];
    }
  }
  for(i=0;i<K;++i){
    B0(i,i) = 1;
  }
  /* Report the result. */

  return Rcpp::List::create(Rcpp::Named("A") = A0,
                            Rcpp::Named("b") = b0,
                            Rcpp::Named("B") = B0,
                            Rcpp::Named("neg_loglik") = fx);

}
