#include "depend_functions.h"

// [[Rcpp::depends(RcppArmadillo)]]

// [[Rcpp::export]]
double loglik_cpp(arma::mat A, arma::mat S, arma::mat response) {
  int N = response.n_rows;
  int J = A.n_rows;
  double ans=0;
  arma::mat temp(J,N);
  arma::mat LS = A * A.t() + S;
  arma::mat LS_multi(J,N);
  arma::vec LS_diag(J);

  for(int i=0;i<J;++i){
    LS_diag(i) = LS(i,i);
    LS(i,i) = 0;
  }

  LS_multi = LS * response.t();
  for(int i=0;i<J;++i){
    for(int j=0;j<N;++j){
      temp(i,j) = LS_diag(i)/2 + LS_multi(i,j);
    }
  }

  for(int i=0;i<N;++i){
    for(int j=0;j<J;++j){
      ans += temp(j,i) * response(i,j) - log(1+exp(temp(j,i)));
    }
  }
  return ans;
}

// [[Rcpp::export]]
double neg_loglik_cpp(arma::mat A, arma::mat S, arma::mat response) {
  int N = response.n_rows;
  double ans=0;
  ans = loglik_cpp(A, S, response);
  return -ans/N;
}

// [[Rcpp::export]]
double loglik_cpp1(arma::mat A, arma::mat S, arma::mat beta, arma::mat Z, arma::mat response) {
  int N = response.n_rows;
  int J = A.n_rows;
  double ans=0;
  arma::mat temp(J,N);
  arma::mat LS = A * A.t() + S;
  arma::mat LS_multi(J,N);
  arma::vec LS_diag(J);
  arma::mat betaz = Z * beta;

  for(int i=0;i<J;++i){
    LS_diag(i) = LS(i,i);
    LS(i,i) = 0;
  }

  LS_multi = LS * response.t();
  for(int i=0;i<J;++i){
    for(int j=0;j<N;++j){
      temp(i,j) = LS_diag(i)/2 + LS_multi(i,j) + betaz(j,i);
    }
  }

  for(int i=0;i<N;++i){
    for(int j=0;j<J;++j){
      ans += temp(j,i) * response(i,j) - log(1+exp(temp(j,i)));
    }
  }
  return ans;
}

// [[Rcpp::export]]
double neg_loglik_cpp1(arma::mat A, arma::mat S, arma::mat beta, arma::mat Z, arma::mat response) {
  int N = response.n_rows;
  double ans=0;
  ans = loglik_cpp1(A, S, beta, Z, response);
  return -ans/N;
}
// [[Rcpp::export]]
double loglik_IRT_cpp(arma::mat A, arma::vec S, arma::mat response) {
  arma::mat Sn = arma::diagmat(S);
  return loglik_cpp(A,Sn,response);
}

// [[Rcpp::export]]
double obj_cpp(arma::mat A, arma::mat S, arma::mat response, double gamma){
  double ans = neg_loglik_cpp(A, S, response);
  S.diag() = arma::zeros(S.n_cols);
  ans += gamma * accu(abs(S));
  return ans;
}

// [[Rcpp::export]]
arma::mat neg_loglik_dev_cpp(arma::mat L, arma::mat S, arma::mat response){
  int J = L.n_rows;
  int N = response.n_rows;
  double temp = 0;
  arma::mat LS = L+S;
  arma::mat LS_multi(J,N);
  arma::vec LS_diag(J);
  arma::mat temp2(J, N);
  arma::vec diag_dev_lik(J);
  arma::mat dev_lik(J,J);
  arma::mat tmpM(J,J);

  for(int i=0;i<J;++i){
    LS_diag(i) = LS(i,i);
    LS(i,i) = 0;
  }

  LS_multi = LS * response.t();
  for(int i=0;i<J;++i){
    for(int j=0;j<N;++j){
      temp = LS_diag(i)/2 + LS_multi(i,j);
      temp2(i, j) = 1/(1+exp(-temp));
    }
  }
  for(int j=0;j<J;++j){
    temp = 0;
    for(int i=0;i<N;++i){
      temp += response(i, j) - temp2(j,i);
    }
    diag_dev_lik(j) = 0.5 * temp;
  }

  tmpM = response.t() * (response - temp2.t());
  dev_lik = -0.5 * (tmpM + tmpM.t()) / N;
  for(int i=0;i<J;++i){
    dev_lik(i,i) = -diag_dev_lik(i) / N;
  }
  return dev_lik;
}


// [[Rcpp::export]]
arma::mat neg_loglik_dev_cpp33(arma::mat L, arma::vec b, arma::mat Y){
  int J = L.n_rows;
  int N = Y.n_rows;
  arma::mat temp = arma::zeros(N,J);
  arma::mat L_off = L;
  arma::mat temp2(N, J);
  arma::mat dev_lik(J,J);
  arma::mat tmpM(J,J);

  for(int i=0;i<J;++i){
    L_off(i,i) = 0;
  }

  temp = 2 * Y * L_off;
  for(int i=0;i<N;++i){
    for(int j=0;j<J;++j){
      temp(i,j) += b(j);
      temp2(i, j) = 1/(1+exp(-temp(i,j)));
    }
  }
  tmpM = 2 * Y.t() * ((Y+1)/2 - temp2);
  dev_lik = -tmpM / N;
  for(int i =0;i<J;++i)
    dev_lik(i,i) = 0;
  return dev_lik;
}
// [[Rcpp::export]]
arma::mat neg_loglik_dev_cpp1(arma::mat L, arma::mat S, arma::mat beta,
                              arma::mat Z, arma::mat response){
  int J = L.n_rows;
  int N = response.n_rows;
  double temp = 0;
  arma::mat LS = L+S;
  arma::mat LS_multi(J,N);
  arma::vec LS_diag(J);
  arma::mat temp2(J, N);
  arma::vec diag_dev_lik(J);
  arma::mat dev_lik(J,J);
  arma::mat tmpM(J,J);
  arma::mat betaz = Z * beta;
  for(int i=0;i<J;++i){
    LS_diag(i) = LS(i,i);
    LS(i,i) = 0;
  }

  LS_multi = LS * response.t();
  for(int i=0;i<J;++i){
    for(int j=0;j<N;++j){
      temp = LS_diag(i)/2 + LS_multi(i,j) + betaz(j,i);
      temp2(i, j) = 1/(1+exp(-temp));
    }
  }
  for(int j=0;j<J;++j){
    temp = 0;
    for(int i=0;i<N;++i){
      temp += response(i, j) - temp2(j,i);
    }
    diag_dev_lik(j) = 0.5 * temp;
  }

  tmpM = response.t() * (response - temp2.t());
  dev_lik = -0.5 * (tmpM + tmpM.t()) / N;
  for(int i=0;i<J;++i){
    dev_lik(i,i) = -diag_dev_lik(i) / N;
  }
  return dev_lik;
}

// [[Rcpp::depends(RcppArmadillo)]]
arma::mat grad_A_cpp(arma::mat A0, arma::mat S0, arma::mat Q, arma::mat response){
  arma::mat L0 = A0 * A0.t();
  arma::mat temp = neg_loglik_dev_cpp(L0, S0, response);
  arma::mat temp_L = 2 * temp * A0;
  temp_L = temp_L % Q;
  return temp_L;
}


arma::mat grad_S_cpp(arma::mat A0, arma::mat S0, arma::mat response){
  return neg_loglik_dev_cpp(A0 * A0.t(), S0, response);
}

arma::mat grad_A_cpp1(arma::mat A0, arma::mat S0, arma::mat Q, arma::mat beta,
                      arma::mat Z, arma::mat response){
  arma::mat temp = neg_loglik_dev_cpp1(A0 * A0.t(), S0, beta, Z, response);
  arma::mat temp_L = 2 * temp * A0;
  temp_L = temp_L % Q;
  return temp_L;
}

arma::mat grad_S_cpp1(arma::mat A0, arma::mat S0, arma::mat beta,
                     arma::mat Z, arma::mat response){
  return neg_loglik_dev_cpp1(A0 * A0.t(), S0, beta, Z, response);
}

arma::mat grad_beta_cpp1(arma::mat A, arma::mat S, arma::mat beta,
                      arma::mat Z, arma::mat response){
  arma::mat L = A * A.t();
  int J = L.n_rows;
  int N = response.n_rows;
  int beta_n = beta.n_rows;
  double temp = 0;
  arma::mat LS = L+S;
  arma::mat LS_multi(J,N);
  arma::vec LS_diag(J);
  arma::mat temp2(J, N);
  arma::mat betaz = Z * beta;
  arma::mat dev_beta(beta_n, J);

  for(int i=0;i<J;++i){
    LS_diag(i) = LS(i,i);
    LS(i,i) = 0;
  }

  LS_multi = LS * response.t();
  for(int i=0;i<J;++i){
    for(int j=0;j<N;++j){
      temp = LS_diag(i)/2 + LS_multi(i,j) + betaz(j,i);
      temp2(i, j) = 1/(1+exp(-temp));
    }
  }
  for(int k=0;k<beta_n;++k){
    for(int j=0;j<J;++j){
      temp = 0;
      for(int i=0;i<N;++i){
        temp += (response(i, j) - temp2(j,i))*Z(i,k);
      }
      dev_beta(k,j) = temp;
    }
  }

  return -dev_beta/N;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
Rcpp::List Update_A_cpp(arma::mat A0, arma::mat S0, arma::mat Q, arma::mat response,
                       double step = 1, double disc = 0.5){
  Rcpp::List ans(2);
  arma::mat grad = grad_A_cpp(A0, S0, Q, response);
  double neg_loglik0 = neg_loglik_cpp(A0, S0, response);

  arma::mat A = A0 - step * grad;
  double neg_loglik1 = neg_loglik_cpp(A, S0, response);
  while(neg_loglik1 > neg_loglik0){
    step = step * disc;
    A = A0 - step * grad;
    neg_loglik1 = neg_loglik_cpp(A, S0, response);
    }
  ans(0) = A;
  ans(1) = step;
  return ans;
}

// [[Rcpp::export]]
double pos_cpp(double x){
  if(x<0)
    x = 0;
  return x;
}
// [[Rcpp::export]]
arma::vec pos_v_cpp(arma::vec x){
  int length = x.n_rows;
  for(int i=0;i<length;++i){
    if(x(i)<0)
      x(i) = 0;
  }
  return x;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
arma::mat soft_thre_cpp(arma::mat x, double mu){
  arma::mat ans = x;
  int x_nrow = x.n_rows;
  int x_ncol = x.n_cols;
  for(int i=0;i<x_nrow;++i){
    for(int j=0;j<x_ncol;++j){
      if(x(i,j)>0)
        ans(i,j) = pos_cpp(x(i,j) - mu);
      else if(x(i,j) == 0)
        ans(i,j) = 0;
      else
        ans(i,j) = -pos_cpp(-x(i,j) - mu);
    }
  }
  return ans;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
double fhat_cpp(arma::mat S, arma::mat S0, arma::mat A0, arma::mat response, double step){
  arma::mat grad = grad_S_cpp(A0, S0, response);
  double ans = neg_loglik_cpp(A0, S0, response);
  arma::mat temp = grad % (S - S0);
  int temp_nrow = temp.n_rows;
  int temp_ncol = temp.n_cols;
  for(int i=0;i<temp_nrow;++i){
    for(int j=0;j<temp_ncol;++j){
      ans += temp(i, j) + pow(S(i,j) - S0(i,j),2)/(2*step);
    }
  }
  return ans;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
Rcpp::List Update_S_cpp(arma::mat A0, arma::mat S0, double gamma, arma::mat response,
                       double step = 1, double disc = 0.5){
  Rcpp::List ans(2);
  arma::mat grad = grad_S_cpp(A0, S0, response);
  arma::mat y = S0 - step * grad;
  arma::mat S = soft_thre_cpp(y, gamma*step);
  int S_nrow = S.n_rows;
  for(int i=0;i<S_nrow;++i){
    S(i,i) = y(i,i);
  }

  double neg_loglik1 = neg_loglik_cpp(A0, S, response);
  double fhat1 = fhat_cpp(S, S0, A0, response, step);
  while(neg_loglik1 > fhat1){
    step = step * disc;
    y = S0 - step * grad;
    S = soft_thre_cpp(y, gamma*step);
    for(int i=0;i<S_nrow;++i){
      S(i,i) = y(i,i);
    }

    neg_loglik1 = neg_loglik_cpp(A0, S, response);
    fhat1 = fhat_cpp(S, S0, A0, response, step);
  }
  ans(0) = S;
  ans(1) = step;
  return ans;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
arma::mat grad_S_supp_cpp(arma::mat A0, arma::mat S0, arma::mat supp, arma::mat response){
  int J = A0.n_rows;
  arma::mat L0 = A0 * A0.t();
  arma::mat tmp = neg_loglik_dev_cpp(L0, S0, response);
  for(int i=0;i<J;++i){
    for(int j=0;j<J;++j){
      tmp(i,j) *= supp(i,j);
    }
  }
  return  tmp;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
Rcpp::List Update_S_supp_cpp(arma::mat A0, arma::mat S0, arma::mat supp, arma::mat response,
                             double step = 1, double disc = 0.5){
  Rcpp::List ans(2);
  arma::mat grad = grad_S_supp_cpp(A0, S0, supp, response);
  double neg_loglik0 = neg_loglik_cpp(A0, S0, response);

  arma::mat S = S0 - step * grad;
  double neg_loglik1 = neg_loglik_cpp(A0, S, response);
  while(neg_loglik1 > neg_loglik0){
    step = step * disc;
    S = S0 - step * grad;
    neg_loglik1 = neg_loglik_cpp(A0, S, response);
  }
  ans(0) = S;
  ans(1) = step;
  return ans;
}
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
arma::mat grad_S_supp_cpp22(arma::mat A0, arma::mat S0, arma::mat B0, arma::mat supp, arma::mat response){
  int J = A0.n_rows;
  arma::mat L0 = A0 * B0 * B0.t() * A0.t();
  arma::mat tmp = neg_loglik_dev_cpp(L0, S0, response);
  for(int i=0;i<J;++i){
    for(int j=0;j<J;++j){
      tmp(i,j) *= supp(i,j);
    }
  }
  return  tmp;
}
// [[Rcpp::depends(RcppArmadillo)]]

// [[Rcpp::export]]
Rcpp::List Update_S_supp_cpp22(arma::mat A0, arma::mat S0, arma::mat B0, arma::mat supp, arma::mat response,
                             double step = 1, double disc = 0.5){
  arma::mat grad = grad_S_supp_cpp22(A0, S0, B0, supp, response);
  double neg_loglik0 = neg_loglik_cpp22(A0, S0, B0, response);

  arma::mat S = S0 - step * grad;
  double neg_loglik1 = neg_loglik_cpp22(A0, S, B0, response);
  while(neg_loglik1 > neg_loglik0){
    step = step * disc;
    S = S0 - step * grad;
    neg_loglik1 = neg_loglik_cpp22(A0, S, B0, response);
  }
  return Rcpp::List::create(Rcpp::Named("S") = S,
                            Rcpp::Named("step") = step);
}


// [[Rcpp::export]]
Rcpp::List reparam1(arma::mat A, arma::mat S, arma::mat Sigma){
  int J = A.n_rows;
  arma::mat L = A*Sigma*(A.t());
  arma::vec e = arma::ones(J);
  arma::vec S_diag = 0.5 * (L + S)*e;
  S = S/4;
  S.diag() = S_diag;
  return Rcpp::List::create(Rcpp::Named("A") = A/2,
                            Rcpp::Named("S") = S);
}

// [[Rcpp::export]]
Rcpp::List reparam2(arma::mat A, arma::mat S, arma::mat Sigma){
  int J = A.n_rows;
  arma::mat L = A*Sigma*(A.t());
  arma::vec e = arma::ones(J);
  arma::vec S_diag = 6*S.diag() -4 * (L + S)*e;
  S = S*4;
  S.diag() = S_diag;
  return Rcpp::List::create(Rcpp::Named("A") = A*2,
                            Rcpp::Named("S") = S);

}
// [[Rcpp::export]]
double par_temp_a_cpp(arma::mat A, arma::vec x, int k, int i, int j){
  if(k==i){
    arma::mat res = x.t() * A.col(j);
    return A(i,j)+res(0)-x(i)*A(i,j);
  }
  else
    return x(i)*A(k,j);
}
// [[Rcpp::export]]
double par2_temp_a_cpp(arma::mat A, arma::mat Sigma, arma::vec y, int k, int i, int j, int it, int jt){
  if(!(i==it)){
    if(k==it)
      return 2*y(i)*Sigma(j,jt);
    if(k==i)
      return 2*y(it)*Sigma(j,jt);
  }
  return 0;
}
// [[Rcpp::export]]
double par_temp_a_cpp_test(arma::mat A, arma::mat Sigma, arma::vec x, int k, int i, int j){
  if(k==i){
    return arma::accu((x.t() * A - x(i) * A.row(i) + A.row(i)) % Sigma.row(j));
  }
  else
    return x(i) * arma::accu(A.row(k) % Sigma.row(j));
}
// [[Rcpp::export]]
double par_temp_a_cpp33(arma::mat A, arma::mat Sigma, arma::vec y, int k, int i, int j){
  if(k==i){
    return 2 * arma::accu((y.t() * A - y(i) * A.row(i)) % Sigma.row(j));
  }
  else
    return 2 * y(i) * arma::accu(A.row(k) % Sigma.row(j));
}
double par_temp_B_cpp(arma::mat A, arma::mat B, arma::vec x, int k, int i, int j){
  arma::mat tmp1 = A(k,i) * A.row(k);
  int J=A.n_rows;
  for(int jj=0;jj<J;++jj){
    if(!(jj==k)){
      tmp1 += x(jj) * (A(jj,i)*A.row(i)+A(k,i)*A.row(jj));
    }
  }
  arma::vec res = tmp1 * B.col(j);
  return(res(0));
}
// [[Rcpp::export]]
double par_temp_b_cpp(int i, int it){
  if(i==it)
    return 0.5;
  else
    return 0;
}
// [[Rcpp::export]]
double par_temp_b_cpp33(int i, int it){
  if(i==it)
    return 1;
  else
    return 0;
}
// [[Rcpp::export]]
double par_temp_beta_cpp(arma::vec Z, int i, int it, int jt){
  if(i==jt)
    return Z(it);
  else
    return 0;
}
// [[Rcpp::export]]
arma::mat calcu_2dev_cpp(arma::mat A, arma::vec b, arma::vec x){
  int J = A.n_rows;
  int K = A.n_cols;
  int i,j,k,it,jt=0;
  arma::mat L=A*A.t();
  arma::mat silly_arma(1,1);
  arma::vec temp = arma::zeros(J);
  for(j=0;j<J;++j){
    temp(j) = (L(j,j)+b(j))/2;
    for(k=0; k<J;++k){
      if(!(k==j))
        temp(j) = temp(j) + L(k,j)*x(k);
    }
  }
  arma::vec temp2 = 1/(1+exp(-temp));
  arma::vec temp3 = exp(temp)/((1+exp(temp))%(1+exp(temp)));
  // calcu 2nd derive matrix
  int row_d2A = J*K-K*(K-1)/2;
  arma::mat d2_A = arma::zeros(row_d2A,row_d2A);
  double container=0;
  int ind1,ind2 =0;
  for(i=0;i<J;++i){
    for(j=0;j<=i&&j<K;++j){
      for(it=0;it<J;++it){
        for(jt=0;jt<=it&&jt<K;++jt){
          silly_arma = x.t()*A.col(j);
          ind1 = j*J+i-j*(j+1)/2;
          ind2 = jt*J+it-jt*(jt+1)/2;
          if(ind1<=ind2){
            d2_A(ind1, ind2) = -temp3(i)*par_temp_a_cpp(A,x,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
            container=0;
            for(k=0;k<J;++k){
              if(!(k==i))
                 container += temp3(k)*par_temp_a_cpp(A,x,k,it,jt)*A(k,j);
            }
            d2_A(ind1, ind2) -= container * x(i);
            if(j==jt){
              if(i==it)
                d2_A(ind1, ind2) += x(i) - temp2(i);
              else
                d2_A(ind1, ind2) += x(i)*(x(it)-temp2(it)) + x(it)*(x(i)-temp2(i));
            }
          }
        }
      }
    }
  }
  arma::mat d2_Ab=arma::zeros(row_d2A, J);
  for(i=0;i<J;++i){
    for(j=0;j<=i&&j<K;++j){
      for(it=0;it<J;++it){
        silly_arma = x.t()*A.col(j);
        ind1 = j*J+i-j*(j+1)/2;
        d2_Ab(ind1,it) = -temp3(i)*par_temp_b_cpp(i,it)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
        for(k=0;k<J;++k){
          if(!(k==i)){
            d2_Ab(ind1,it) -= temp3(k)*par_temp_b_cpp(k,it)*x(i)*A(k,j);
          }
        }
      }
    }
  }

   arma::mat d2_b=arma::zeros(J,J);
  for(i=0;i<J;++i){
    for(it=0;it<J;++it){
      if(i==it) d2_b(i,it) = -0.25*temp3(i);
      else d2_b(i,it) = 0;
    }
  }

  arma::mat H = arma::zeros(J*(K+1)-K*(K-1)/2, J*(K+1)-K*(K-1)/2);
  ind1 = row_d2A+J;
  H.submat(0,0,row_d2A-1,row_d2A-1) = d2_A;
  H.submat(0,row_d2A,row_d2A-1,ind1-1) = d2_Ab;
  H.submat(row_d2A, row_d2A, ind1-1, ind1-1) = d2_b;
  return H;
}

// [[Rcpp::export]]
arma::mat calcu_2dev_cov_cpp(arma::mat A, arma::vec b,arma::mat beta, arma::vec x,arma::vec Z){
  int J = A.n_rows;
  int K = A.n_cols;
  int beta_n = Z.n_rows;
  int i,j,k,it,jt=0;
  arma::mat L=A*A.t();
  arma::mat silly_arma(1,1);
  arma::vec temp = arma::zeros(J);
  for(j=0;j<J;++j){
    temp(j) = (L(j,j)+b(j))/2;
    for(k=0; k<J;++k){
      if(!(k==j))
        temp(j) = temp(j) + L(k,j)*x(k);
    }
    silly_arma = beta.col(j).t() * Z;
    temp(j) = temp(j) + silly_arma(0);
  }
  arma::vec temp2 = 1/(1+exp(-temp));
  arma::vec temp3 = exp(temp)/((1+exp(temp))%(1+exp(temp)));
  // calcu 2nd derive matrix
  int row_d2A = J*K-K*(K-1)/2;
  arma::mat d2_A = arma::zeros(row_d2A,row_d2A);
  double container=0;
  int ind1,ind2 =0;
  for(i=0;i<J;++i){
    for(j=0;j<=i&&j<K;++j){
      for(it=0;it<J;++it){
        for(jt=0;jt<=it&&jt<K;++jt){
          silly_arma = x.t()*A.col(j);
          ind1 = j*J+i-j*(j+1)/2;
          ind2 = jt*J+it-jt*(jt+1)/2;
          if(ind1<=ind2){
            d2_A(ind1, ind2) = -temp3(i)*par_temp_a_cpp(A,x,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
            container=0;
            for(k=0;k<J;++k){
              if(!(k==i))
                container += temp3(k)*par_temp_a_cpp(A,x,k,it,jt)*A(k,j);
            }
            d2_A(ind1, ind2) -= container * x(i);
            if(j==jt){
              if(i==it)
                d2_A(ind1, ind2) += x(i) - temp2(i);
              else
                d2_A(ind1, ind2) += x(i)*(x(it)-temp2(it)) + x(it)*(x(i)-temp2(i));
            }
          }
        }
      }
    }
  }
  arma::mat d2_Ab=arma::zeros(row_d2A, J);
  for(i=0;i<J;++i){
    for(j=0;j<=i&&j<K;++j){
      for(it=0;it<J;++it){
        silly_arma = x.t()*A.col(j);
        ind1 = j*J+i-j*(j+1)/2;
        d2_Ab(ind1,it) = -temp3(i)*par_temp_b_cpp(i,it)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
        for(k=0;k<J;++k){
          if(!(k==i)){
            d2_Ab(ind1,it) -= temp3(k)*par_temp_b_cpp(k,it)*x(i)*A(k,j);
          }
        }
      }
    }
  }
  arma::mat d2_Abeta=arma::zeros(row_d2A,beta_n*J);
  for(i=0;i<J;++i){
    for(j=0;j<=i&&j<K;++j){
      for(it=0;it<beta_n;++it){
        for(jt=0;jt<J;++jt){
          ind1 = j*J+i-j*(j+1)/2;
          ind2 = jt*beta_n+it;
          silly_arma = x.t()*A.col(j);
          d2_Abeta(ind1,ind2) = -temp3(i)*par_temp_beta_cpp(Z,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
          for(k=0;k<J;++k){
            if(!(k==i)){
              d2_Abeta(ind1,ind2) -= temp3(k)*par_temp_beta_cpp(Z,k,it,jt)*x(i)*A(k,j);
            }
          }
        }
      }
    }
  }
  arma::mat d2_b=arma::zeros(J,J);
  for(i=0;i<J;++i){
    for(it=0;it<J;++it){
      if(i==it) d2_b(i,it) = -0.25*temp3(i);
      else d2_b(i,it) = 0;
    }
  }
  arma::mat d2_bbeta=arma::zeros(J, beta_n*J);
  for(i=0;i<J;++i){
    for(it=0;it<beta_n;++it){
      for(jt=0;jt<J;++jt){
        ind2 = jt*beta_n+it;
        if(i==jt)
          d2_bbeta(i,ind2) = -0.5*temp3(i)*Z(it);
        else
          d2_bbeta(i,ind2) =0;
      }
    }
  }
  arma::mat d2_beta=arma::zeros(beta_n*J, beta_n*J);
  for(i=0;i<beta_n;++i){
    for(j=0;j<J;++j){
      for(it=0;it<beta_n;++it){
        for(jt=0;jt<J;++jt){
          ind1 = j*beta_n + i;
          ind2 = jt*beta_n+it;
          if(j==jt)
            d2_beta(ind1, ind2) = -Z(i)*temp3(j)*Z(it);
          else
            d2_beta(ind1, ind2) = 0;
        }
      }
    }
  }
  arma::mat H = arma::zeros(J*(K+1+beta_n)-K*(K-1)/2, J*(K+1+beta_n)-K*(K-1)/2);
  ind1 = row_d2A+J;
  ind2 = row_d2A+J+beta_n*J;
  H.submat(0,0,row_d2A-1,row_d2A-1) = d2_A;
  H.submat(0,row_d2A,row_d2A-1,ind1-1) = d2_Ab;
  H.submat(0,ind1, row_d2A-1,ind2-1) = d2_Abeta;
  H.submat(row_d2A, row_d2A, ind1-1, ind1-1) = d2_b;
  H.submat(row_d2A, ind1, ind1-1, ind2-1) = d2_bbeta;
  H.submat(ind1, ind1, ind2-1,ind2-1) = d2_beta;
  // H.submat(row_d2A, 0,ind1-1,row_d2A-1) = d2_Ab.t();
  // H.submat(ind1,0,ind2-1,row_d2A-1) = d2_Abeta.t();
  // H.submat(ind1,row_d2A, ind2-1, ind1-1) = d2_bbeta.t();
  return H;
}

// [[Rcpp::export]]
arma::mat calcu_2dev_cov_new_cpp(arma::mat A, arma::vec b, arma::mat Q, arma::mat beta, arma::vec x,arma::vec Z){
  int J = A.n_rows;
  int K = A.n_cols;
  int beta_n = Z.n_rows;
  int i,j,k,it,jt=0;
  arma::mat L=A*A.t();
  arma::mat silly_arma(1,1);
  arma::vec temp = arma::zeros(J);
  for(j=0;j<J;++j){
    temp(j) = (L(j,j)+b(j))/2;
    for(k=0; k<J;++k){
      if(!(k==j))
        temp(j) = temp(j) + L(k,j)*x(k);
    }
    silly_arma = beta.col(j).t() * Z;
    temp(j) = temp(j) + silly_arma(0);
  }
  arma::vec temp2 = 1/(1+exp(-temp));
  arma::vec temp3 = exp(temp)/((1+exp(temp))%(1+exp(temp)));
  // calcu 2nd derive matrix
  int row_d2A = arma::accu(Q);
  arma::mat d2_A = arma::zeros(row_d2A,row_d2A);
  double container=0;
  int ind1=0;
  int ind2 =0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      if(Q(i,j)>0){
        ind2 = 0;
        for(it=0;it<J;++it){
          for(jt=0;jt<K;++jt){
            if(Q(it,jt)>0){
              silly_arma = x.t()*A.col(j);
              if(ind1<=ind2){
                d2_A(ind1, ind2) = -temp3(i)*par_temp_a_cpp(A,x,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
                container=0;
                for(k=0;k<J;++k){
                  if(!(k==i))
                    container += temp3(k)*par_temp_a_cpp(A,x,k,it,jt)*A(k,j);
                }
                d2_A(ind1, ind2) -= container * x(i);
                if(j==jt){
                  if(i==it)
                    d2_A(ind1, ind2) += x(i) - temp2(i);
                  else
                    d2_A(ind1, ind2) += x(i)*(x(it)-temp2(it)) + x(it)*(x(i)-temp2(i));
                }
              }
              ind2 += 1;
            }
          }
        }
        ind1 += 1;
      }
    }
  }
  arma::mat d2_Ab=arma::zeros(row_d2A, J);
  ind1 = 0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      if(Q(i,j)>0){
        for(it=0;it<J;++it){
          silly_arma = x.t()*A.col(j);
          d2_Ab(ind1,it) = -temp3(i)*par_temp_b_cpp(i,it)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
          for(k=0;k<J;++k){
            if(!(k==i)){
              d2_Ab(ind1,it) -= temp3(k)*par_temp_b_cpp(k,it)*x(i)*A(k,j);
            }
          }
        }
        ind1 += 1;
      }
    }
  }
  arma::mat d2_Abeta=arma::zeros(row_d2A,beta_n*J);
  ind1 = 0;
  ind2 = 0;
  for(i=0;i<J;++i){
    for(j=0;j<K;++j){
      if(Q(i,j)>0){
        for(it=0;it<beta_n;++it){
          for(jt=0;jt<J;++jt){
            ind2 = jt*beta_n+it;
            silly_arma = x.t()*A.col(j);
            d2_Abeta(ind1,ind2) = -temp3(i)*par_temp_beta_cpp(Z,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
            for(k=0;k<J;++k){
              if(!(k==i)){
                d2_Abeta(ind1,ind2) -= temp3(k)*par_temp_beta_cpp(Z,k,it,jt)*x(i)*A(k,j);
              }
            }
          }
        }
        ind1 += 1;
      }
    }
  }
  arma::mat d2_b=arma::zeros(J,J);
  for(i=0;i<J;++i){
    for(it=0;it<J;++it){
      if(i==it) d2_b(i,it) = -0.25*temp3(i);
      else d2_b(i,it) = 0;
    }
  }
  arma::mat d2_bbeta=arma::zeros(J, beta_n*J);
  for(i=0;i<J;++i){
    for(it=0;it<beta_n;++it){
      for(jt=0;jt<J;++jt){
        ind2 = jt*beta_n+it;
        if(i==jt)
          d2_bbeta(i,ind2) = -0.5*temp3(i)*Z(it);
        else
          d2_bbeta(i,ind2) =0;
      }
    }
  }
  arma::mat d2_beta=arma::zeros(beta_n*J, beta_n*J);
  for(i=0;i<beta_n;++i){
    for(j=0;j<J;++j){
      for(it=0;it<beta_n;++it){
        for(jt=0;jt<J;++jt){
          ind1 = j*beta_n + i;
          ind2 = jt*beta_n+it;
          if(j==jt)
            d2_beta(ind1, ind2) = -Z(i)*temp3(j)*Z(it);
          else
            d2_beta(ind1, ind2) = 0;
        }
      }
    }
  }
  arma::mat H = arma::zeros(arma::accu(Q)+J+beta_n*J, arma::accu(Q)+J+beta_n*J);
  ind1 = row_d2A+J;
  ind2 = row_d2A+J+beta_n*J;
  H.submat(0,0,row_d2A-1,row_d2A-1) = d2_A;
  H.submat(0,row_d2A,row_d2A-1,ind1-1) = d2_Ab;
  H.submat(0,ind1, row_d2A-1,ind2-1) = d2_Abeta;
  H.submat(row_d2A, row_d2A, ind1-1, ind1-1) = d2_b;
  H.submat(row_d2A, ind1, ind1-1, ind2-1) = d2_bbeta;
  H.submat(ind1, ind1, ind2-1,ind2-1) = d2_beta;
  // H.submat(row_d2A, 0,ind1-1,row_d2A-1) = d2_Ab.t();
  // H.submat(ind1,0,ind2-1,row_d2A-1) = d2_Abeta.t();
  // H.submat(ind1,row_d2A, ind2-1, ind1-1) = d2_bbeta.t();
  return H;
}

// [[Rcpp::export]]
arma::mat calcu_2dev_cpp_new(arma::mat A, arma::vec b, arma::mat Q, arma::vec x){
  int J = A.n_rows;
  int K = A.n_cols;
  int i,j,k,it,jt=0;
  arma::mat L=A*A.t();
  arma::mat silly_arma(1,1);
  arma::vec temp = arma::zeros(J);
  for(j=0;j<J;++j){
    temp(j) = (L(j,j)+b(j))/2;
    for(k=0; k<J;++k){
      if(!(k==j))
        temp(j) = temp(j) + L(k,j)*x(k);
    }
  }
  arma::vec temp2 = 1/(1+exp(-temp));
  arma::vec temp3 = exp(temp)/((1+exp(temp))%(1+exp(temp)));
  // calcu 2nd derive matrix
  int row_d2A = arma::accu(Q);
  arma::mat d2_A = arma::zeros(row_d2A,row_d2A);
  double container=0;
  int ind1=0;
  int ind2 =0;
  for(j=0;j<K;++j){
    for(i=0;i<J;++i){
      if(Q(i,j)>0){
        ind2 = 0;
        silly_arma = x.t()*A.col(j);
        for(jt=0;jt<K;++jt){
          for(it=0;it<J;++it){
            if(Q(it,jt)>0){
              if(ind1<=ind2){
                d2_A(ind1, ind2) = -temp3(i)*par_temp_a_cpp(A,x,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
                container=0;
                for(k=0;k<J;++k){
                  if(!(k==i))
                    container += temp3(k)*par_temp_a_cpp(A,x,k,it,jt)*A(k,j);
                }
                d2_A(ind1, ind2) -= container * x(i);
                if(j==jt){
                  if(i==it)
                    d2_A(ind1, ind2) += x(i) - temp2(i);
                  else
                    d2_A(ind1, ind2) += x(i)*(x(it)-temp2(it)) + x(it)*(x(i)-temp2(i));
                }
                //Rprintf("i:%d, j:%d, it:%d, jt:%d, ind1:%d, ind2:%d\n",i,j,it,jt,ind1,ind2);
              }
              ind2 += 1;
            }
          }
        }
        ind1 += 1;
      }
    }
  }
  arma::mat d2_Ab=arma::zeros(row_d2A, J);
  ind1 = 0;
  for(j=0;j<K;++j){
    for(i=0;i<J;++i){
      if(Q(i,j)){
        for(it=0;it<J;++it){
          silly_arma = x.t()*A.col(j);
          // ind1 = j*J+i-j*(j+1)/2;
          d2_Ab(ind1,it) = -temp3(i)*par_temp_b_cpp(i,it)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
          for(k=0;k<J;++k){
            if(!(k==i)){
              d2_Ab(ind1,it) -= temp3(k)*par_temp_b_cpp(k,it)*x(i)*A(k,j);
            }
          }
        }
        ind1 += 1;
      }
    }
  }

  arma::mat d2_b=arma::zeros(J,J);
  for(i=0;i<J;++i){
    for(it=0;it<J;++it){
      if(i==it) d2_b(i,it) = -0.25*temp3(i);
      else d2_b(i,it) = 0;
    }
  }

  arma::mat H = arma::zeros(arma::accu(Q)+J, arma::accu(Q)+J);
  ind1 = row_d2A+J;
  H.submat(0,0,row_d2A-1,row_d2A-1) = d2_A;
  H.submat(0,row_d2A,row_d2A-1,ind1-1) = d2_Ab;
  H.submat(row_d2A, row_d2A, ind1-1, ind1-1) = d2_b;
  return H;
}

// [[Rcpp::export]]
arma::mat calcu_2dev_cpp_new1(arma::mat A, arma::vec b, arma::mat QN, arma::vec x){
  int J = A.n_rows;
  int K = A.n_cols;
  int i,j,k,it,jt=0;
  arma::mat L=A*A.t();
  arma::mat silly_arma(1,1);
  arma::vec temp = arma::zeros(J);
  for(j=0;j<J;++j){
    temp(j) = (L(j,j)+b(j))/2;
    for(k=0; k<J;++k){
      if(!(k==j))
        temp(j) = temp(j) + L(k,j)*x(k);
    }
  }
  arma::vec temp2 = 1/(1+exp(-temp));
  arma::vec temp3 = exp(temp)/((1+exp(temp))%(1+exp(temp)));
  // calcu 2nd derive matrix
  int row_d2A = QN.n_rows;
  arma::mat d2_A = arma::zeros(row_d2A,row_d2A);
  double container=0;
  int ind1=0;
  int ind2 =0;
  for(ind1=0;ind1<row_d2A;++ind1){
    for(ind2=0;ind2<row_d2A;++ind2){
      if(ind1<=ind2){
        i = QN(ind1,0);
        j = QN(ind1,1);
        it = QN(ind2,0);
        jt = QN(ind2,1);
        silly_arma = x.t()*A.col(j);
        d2_A(ind1, ind2) = -temp3(i)*par_temp_a_cpp(A,x,i,it,jt)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
        container=0;
        for(k=0;k<J;++k){
          if(!(k==i))
            container += temp3(k)*par_temp_a_cpp(A,x,k,it,jt)*A(k,j);
        }
        d2_A(ind1, ind2) -= container * x(i);
        if(j==jt){
          if(i==it)
            d2_A(ind1, ind2) += x(i) - temp2(i);
          else
            d2_A(ind1, ind2) += x(i)*(x(it)-temp2(it)) + x(it)*(x(i)-temp2(i));
        }
      }
    }
  }

  arma::mat d2_Ab=arma::zeros(row_d2A, J);
  ind1 = 0;
  for(ind1=0;ind1<row_d2A; ++ind1){
    i = QN(ind1,0);
    j = QN(ind1,1);
    for(it=0;it<J;++it){
      silly_arma = x.t()*A.col(j);
      d2_Ab(ind1,it) = -temp3(i)*par_temp_b_cpp(i,it)*(A(i,j)+silly_arma(0)-x(i)*A(i,j));
      for(k=0;k<J;++k){
        if(!(k==i)){
          d2_Ab(ind1,it) -= temp3(k)*par_temp_b_cpp(k,it)*x(i)*A(k,j);
        }
      }
    }
  }
  arma::mat d2_b=arma::zeros(J,J);
  for(i=0;i<J;++i){
    for(it=0;it<J;++it){
      if(i==it) d2_b(i,it) = -0.25*temp3(i);
      else d2_b(i,it) = 0;
    }
  }
  arma::mat H = arma::zeros(row_d2A+J, row_d2A+J);
  ind1 = row_d2A+J;
  H.submat(0,0,row_d2A-1,row_d2A-1) = d2_A;
  H.submat(0,row_d2A,row_d2A-1,ind1-1) = d2_Ab;
  H.submat(row_d2A, row_d2A, ind1-1, ind1-1) = d2_b;
  return H;
}

// [[Rcpp::export]]
arma::mat calcu_2dev_cpp_new_test(arma::mat A, arma::vec b, arma::mat Sigma, arma::mat QN, arma::vec y){
  int J = A.n_rows;
  int K = A.n_cols;
  int i,j,k,it,jt=0;
  int k1,k2;
  arma::mat L=A*Sigma*A.t();
  arma::mat silly_arma(1,1);
  arma::vec temp = arma::zeros(J);
  for(j=0;j<J;++j){
    temp(j) = b(j);
    for(k=0; k<J;++k){
      if(!(k==j))
        temp(j) = temp(j) + 2 * L(k,j)*y(k);
    }
  }
  arma::vec temp2 = 1/(1+exp(-temp));
  arma::vec temp3 = exp(temp)/((1+exp(temp))%(1+exp(temp)));
  arma::mat temp4 = arma::zeros(J, J);
  for(i=0;i<J;++i){
    for(j=0;j<J;++j){
      if(!(i==j))
        temp4(i,j) = 2 * ((y(j)+1)/2-temp2(j))*y(i);
      else
        temp4(i,j) = 0;
    }
  }
  // calcu 2nd derive matrix
  int row_d2A = QN.n_rows;
  arma::mat d2_A = arma::zeros(row_d2A,row_d2A);
  double container=0;
  int ind1=0;
  int ind2 =0;
  for(ind1=0;ind1<row_d2A;++ind1){
    for(ind2=0;ind2<row_d2A;++ind2){
      if(ind1<=ind2){
        i = QN(ind1,0);
        j = QN(ind1,1);
        it = QN(ind2,0);
        jt = QN(ind2,1);
        d2_A(ind1, ind2) = -temp3(i) * par_temp_a_cpp33(A,Sigma, y, i, it, jt) * 2 *arma::accu((y.t() * A - y(i)*A.row(i))*Sigma.col(j));
        container=0;
        for(k=0;k<J;++k){
          if(k!=i)
            container += temp3(k)*par_temp_a_cpp33(A,Sigma, y, k, it, jt) * 2 * arma::accu(A.row(k) % Sigma.row(j));
        }
        d2_A(ind1, ind2) -= container * y(i);
        if(i!=it)
          d2_A(ind1, ind2) += (temp4(i,it) + temp4(it,i))*Sigma(j,jt);
      }
    }
  }
  arma::mat d2_Ab=arma::zeros(row_d2A, J);
  ind1 = 0;
  for(ind1=0;ind1<row_d2A; ++ind1){
    i = QN(ind1,0);
    j = QN(ind1,1);
    for(it=0;it<J;++it){
      if(it != i)
        d2_Ab(ind1, it) = -2 * temp3(it) * y(i) * arma::accu(A.row(it) * Sigma.col(j));
      else
        d2_Ab(ind1, it) = -2*temp3(it) * arma::accu((y.t()*A-y(it)*A.row(it))*Sigma.col(j));
    }
  }
  // arma::mat d2_AB=arma::zeros(row_d2A, K*(K-1)/2);
  // ind1 = 0;
  // ind2 = 0;
  // for(ind1 =0;ind1<row_d2A; ++ind1){
  //   i = QN(ind1,0);
  //   j = QN(ind1,1);
  //   ind2 = 0;
  //   for(it=1;it<K;++it){
  //     for(jt=0;jt<it;++jt){
  //       d2_AB(ind1, ind2) = -temp3(i) * par_temp_B_cpp(A,B,x,i,it,jt) * arma::accu((x.t()*A-x(i)*A.row(i)+A.row(i))%Sigma.row(j));
  //       for(k=0;k<J;++k){
  //         if(!(k==i)){
  //           d2_AB(ind1, ind2) -= temp3(k)*par_temp_B_cpp(A,B,x,k,it,jt)*x(i)*arma::accu(A.row(k) % Sigma.row(j));
  //         }
  //       }
  //       arma::mat tmp21 = 2*temp4.row(i)*A;
  //       if(j==it){
  //         silly_arma = tmp21 * B.col(jt);
  //         d2_AB(ind1, ind2) += silly_arma(0);
  //       }
  //       else{
  //         d2_AB(ind1, ind2) += tmp21(it) * B(j,jt);
  //       }
  //       ind2 += 1;
  //     }
  //   }
  // }
  arma::mat d2_b=arma::zeros(J,J);
  for(i=0;i<J;++i){
    for(it=0;it<J;++it){
      if(i==it)
        d2_b(i,it) = -temp3(i);
      else
        d2_b(i,it) = 0;
    }
  }
  // arma::mat d2_bB= arma::zeros(J, K*(K-1)/2);
  // for(i=0;i<J;++i){
  //   ind2 = 0;
  //   for(it=1;it<K;++it){
  //     for(jt=0;jt<it;++jt){
  //       d2_bB(i,ind2) = -0.5 * temp3(i) * par_temp_B_cpp(A,B,x,i,it,jt);
  //       ind2 += 1;
  //     }
  //   }
  // }
  // arma::mat d2_BB = arma::zeros(K*(K-1)/2, K*(K-1)/2);
  // ind1=0;
  // ind2=0;
  // for(i=1;i<K;++i){
  //   for(j=0;j<i;++j){
  //     ind2 = 0;
  //     for(it=1;it<K;++it){
  //       for(jt=0;jt<it;++jt){
  //         for(k=0;k<J;++k){
  //           d2_BB(ind1, ind2) += -temp3(k)*par_temp_B_cpp(A,B,x,k,i,j)*par_temp_B_cpp(A,B,x,k,it,jt);
  //         }
  //         if(j==jt){
  //           for(k=0;k<J;++k){
  //             d2_BB(ind1,ind2) += (x(k)-temp2(k))*(arma::accu(x.t() * A.col(i) * A.row(k) + A(k,i) * x.t() * A)-
  //               2*x(k)*A(k,i)*A(k,it)+A(k,i)*A(k,it));
  //           }
  //         }
  //       }
  //     }
  //   }
  // }
  // for(i=1;i<K;++i){
  //   for(j=0;j<i;++j){
  //     ind2=0;
  //     for(it=1;it<K;++it){
  //       for(jt=0;jt<it;++jt){
  //
  //         for(k1=0;k1<J;++k1){
  //           for(k2=0;k2<J;++k2){
  //             if(k1 == k2){
  //               d2_BB(ind1,ind2) += -0.5 * temp3(k1)*par_temp_B_cpp(A,B,x,k1,it,jt)*arma::accu((A(k1,i)*A.row(k2)+A(k2,i)*A.row(k1))%B.col(j).t());
  //             }
  //             else{
  //               d2_BB(ind1, ind2) += -0.5 *(temp3(k1)*par_temp_B_cpp(A,B,x,k1,it,jt)*x(k2)+
  //                 temp3(k2)*par_temp_B_cpp(A,B,x,k2,it,jt)*x(k1))*arma::accu((A(k1,i)*A.row(k2)+A(k2,i)*A.row(k1))%B.col(j).t());
  //             }
  //           }
  //         }
  //         if(j==jt){
  //           d2_BB(ind1,ind2) += arma::accu(temp4 % (A.col(i)*A.col(it).t() + A.col(it)*A.col(i).t()));
  //         }
  //         ind2 += 1;
  //       }
  //     }
  //     ind1 += 1;
  //   }
  // }

  ind1 = row_d2A+J;
  arma::mat H = arma::zeros(ind1, ind1);
  H.submat(0,0,row_d2A-1,row_d2A-1) = d2_A;
  H.submat(0,row_d2A,row_d2A-1,ind1-1) = d2_Ab;
  H.submat(row_d2A, row_d2A, ind1-1, ind1-1) = d2_b;
  return H;
}


// [[Rcpp::export]]
double neg_loglik_cpp22(arma::mat A, arma::mat S, arma::mat B, arma::mat response) {
  int N = response.n_rows;
  int J = A.n_rows;
  double ans=0;
  arma::mat temp(J,N);
  arma::mat LS = A * B * B.t() * A.t() + S;
  arma::mat LS_multi(J,N);
  arma::vec LS_diag(J);

  for(int i=0;i<J;++i){
    LS_diag(i) = LS(i,i);
    LS(i,i) = 0;
  }

  LS_multi = LS * response.t();
  for(int i=0;i<J;++i){
    for(int j=0;j<N;++j){
      temp(i,j) = LS_diag(i)/2 + LS_multi(i,j);
    }
  }

  for(int i=0;i<N;++i){
    for(int j=0;j<J;++j){
      ans += temp(j,i) * response(i,j) - log(1+exp(temp(j,i)));
    }
  }
  return -ans/N;
}


// [[Rcpp::export]]
arma::mat grad_A_cpp22(arma::mat A0, arma::mat S0, arma::mat B0, arma::mat Q, arma::mat response){
  arma::mat L0 = A0 * B0 * B0.t() * A0.t();
  arma::mat temp = neg_loglik_dev_cpp(L0, S0, response);
  arma::mat temp_L = 2 * temp * A0 * B0 * B0.t();
  temp_L = temp_L % Q;
  return temp_L;
}

// [[Rcpp::export]]
arma::mat grad_A_cpp33(arma::mat A, arma::vec b, arma::mat Sigma, arma::mat Y){
  arma::mat L = A * Sigma * A.t();
  arma::mat temp = neg_loglik_dev_cpp33(L, b, Y);
  arma::mat temp_L = temp * A * Sigma + temp.t() * A * Sigma;
  return temp_L;
}

// [[Rcpp::export]]
arma::mat grad_b_cpp333(arma::mat A, arma::vec b, arma::mat Sigma, arma::mat Y){
  int N = Y.n_rows;
  int J = Y.n_cols;
  arma::mat L = A * Sigma * A.t();
  arma::mat L_off = L;
  for(int i=0;i<J;++i) L_off(i,i) = 0;
  arma::mat temp = 2 * Y * L_off;
  temp = temp + arma::ones(N) * b.t();
  temp = 1/(1+exp(-temp));
  return -1/N * ((Y+1)/2 - temp).t() * arma::ones(N);
  // N <- nrow(Y)
  // L = A %*% Sigma %*% t(A)
  // L_off = L
  // diag(L_off) = 0
  // temp = 2 * Y %*% L_off
  // temp = temp + t(rep(b,N))
  // temp = 1/(1+exp(-temp))
}

// [[Rcpp::export]]
arma::mat grad_S_cpp22(arma::mat A0, arma::mat S0, arma::mat B0, arma::mat response){
  return neg_loglik_dev_cpp(A0 * B0 * B0.t() * A0.t(), S0, response);
}


// [[Rcpp::export]]
arma::mat grad_B_cpp22(arma::mat A0, arma::mat S0, arma::mat B0, arma::mat response){
  arma::mat temp = neg_loglik_dev_cpp(A0 * B0 * B0.t() * A0.t(), S0, response);
  return 2 * A0.t() * temp * A0 * B0;
}



