#include <RcppArmadillo.h>
#include "lbfgs.h"

static lbfgsfloatval_t evaluate(
    void *instance,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step
)
{
  int k =0;
  int i=0;
  lbfgsfloatval_t fx = 0.0;
  lbfgsfloatval_t *params = (lbfgsfloatval_t *)instance;
  const int J = params[k++];
  const int N = params[k++];
  arma::vec K0j(J);
  arma::mat response(N,J);

  for(int i=0;i<J;++i)
    K0j(i) = params[k++];

  const int j = params[k++];
  for(int p=0;p<N;++p){
    for(int q=0;q<J;++q){
      response(p,q) = params[k++];
    }
  }
  double lambda = params[k++];

  arma::vec temp(N);

  for(i=0;i<N;++i){
    temp(i) = 0.5 * x[j];
    for(k=0;k<J;++k){
      if(k != j)
        temp(i) += response(i,k) * x[k];
    }
  }
  for(int i=0;i<N;++i){
    fx += response(i,j) * temp(i) - log(1+exp(temp(i)));
  }
  fx = -lambda * fx / N;
  for(i=0;i<J;++i){
    fx += pow((x[i]-K0j(i)),2)/2;
  }

  arma::vec temp2 = 1/(1+exp(-temp));
  for(i=0;i<J;++i){
    if(i == j)
      g[i] = 0.5 * arma::accu(response.col(i) - temp2);
    else
      g[i] = arma::accu(response.col(i) % (response.col(j) - temp2));
  }
  for(i=0;i<J;++i){
    g[i] = -lambda/N * g[i] + x[i] - K0j(i);
  }
  return fx;
}

// [[Rcpp::export]]
arma::vec proximal_f_Kj5(arma::vec K0j, int j, arma::mat response, double lambda)
{
  int i,k,m,n, ret = 0;
  int N = response.n_rows;
  int J = response.n_cols;

  lbfgsfloatval_t fx;
  lbfgsfloatval_t *x = lbfgs_malloc(J);
//  lbfgs_parameter_t param;
  lbfgsfloatval_t *params = lbfgs_malloc(J+N*J+10);
  k=0;
  params[k++] = J;
  params[k++] = N;
  for(i=0;i<J;++i){
    params[k++] = K0j(i);
  }
  params[k++] = j-1;
  for(m =0;m<N;++m){
    for(n=0;n<J;++n){
      params[k++] = response(m,n);
    }
  }
  params[k++] = lambda;

  /* Initialize the variables. */
  for (i = 0;i < J;++i) {
    x[i] = K0j(i);
  }
  /* Initialize the parameters for the L-BFGS optimization. */
//  lbfgs_parameter_init(&param);
  /*param.linesearch = LBFGS_LINESEARCH_BACKTRACKING;*/
  /*
   Start the L-BFGS optimization; this will invoke the callback functions
   evaluate() and progress() when necessary.
   */
  ret = lbfgs(J, x, &fx, evaluate, NULL, params, NULL);
  /* Report the result. */
//  printf("L-BFGS optimization terminated with status code = %d\n", ret);
//  printf("  fx = %f, x[0] = %f, x[1] = %f\n", fx, x[0], x[1]);
  arma::vec ans(J);
  for(int i=0;i<J;++i)
    ans(i) = x[i];
  lbfgs_free(x);
  lbfgs_free(params);
  return ans;
}



